/*
 * Copyright 1990 MCNC and Vivek Khera
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of MCNC not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  MCNC makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * MCNC DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCNC
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Vivek Khera, MCNC
 */

/*
 * Access Control List extension for X11R4. intended use is for building up
 * an ACL which a kerberos-based authorization scheme can use to allow or
 * disallow access to the server.
 *
 * V. Khera	25-JUL-1990 
 */

/* $Header: kacl.c,v 1.7 90/09/18 10:27:49 bin Exp $ */

#include "X.h"
#include "Xproto.h"
#include "misc.h"
#include "os.h"
#include "dixstruct.h"
#include "extnsionst.h"
#define _KACL_SERVER_
#include "kaclstr.h"

static unsigned char KACLReqCode;

typedef struct _ACLEntry {
	char *name;
	struct _ACLEntry *next;
} ACLEntry;

static ACLEntry *listOfNames = NULL;	/* linked list of allowed names */

void
KACLExtensionInit()
{
	ExtensionEntry *extEntry, *AddExtension();
	static int      ProcKACLDispatch(), SProcKACLDispatch();
	static void     KACLResetProc();

	if (extEntry = AddExtension(KACLNAME, 0, 0,
				    ProcKACLDispatch, SProcKACLDispatch,
				    KACLResetProc, StandardMinorOpcode))
		KACLReqCode = (unsigned char) extEntry->base;
}

/* 
 * restore server to state prepared to init extension once again. 
 * listOfNames must be set to NULL so that server will use the 
 * fall-back scheme of the magic-cookie.
 */
static void
KACLResetProc(extEntry)
ExtensionEntry *extEntry;
{
	/* chase down list and deallocate used memory */
	ACLEntry *cur = listOfNames;
	ACLEntry *next;

	while (cur) {
		xfree(cur->name);
		next = cur->next;
		xfree(cur);
		cur = next;
	}

	listOfNames = NULL;
}

/* 
 * add a user to the acl. this code is modelled after the code to add 
 * to the font path
 */
static int
ProcKACLChangeUsers(client)
register ClientPtr client;
{
	REQUEST(xKACLChangeUsersReq);
	int result;
	unsigned long n, total;
	unsigned char *ptr;
	static int AddUser(), DelUser();

	REQUEST_AT_LEAST_SIZE(xKACLChangeUsersReq);

	total = (stuff->length << 2) - sizeof(xKACLChangeUsersReq);
	ptr = (unsigned char *)&stuff[1];	/* length byte */
	n = (*ptr + 1);

	if ((total == 0) || (total < n) || (total - n >= 4))
		return(BadLength);

	if (stuff->mode == aclUserInsert) {
		result = AddUser(client, ptr + 1);
	} else if (stuff->mode == aclUserDelete) {
		result = DelUser(client, ptr + 1);
	} else {
		client->errorValue = stuff->mode;
		return BadValue;
	}

	if (!result)
		result = client->noClientException;

	return result;
}

static int
ProcKACLGetUsers(client)
register ClientPtr client;
{
	REQUEST(xReq);
	xKACLGetUsersReply rep;
	int n;
	char *nameList;
	static char *GetUserList();

	REQUEST_SIZE_MATCH(xReq);

	nameList = GetUserList();	/* colon separated list as C string */
	if (!nameList)
		return (BadAlloc);

	rep.type = X_Reply;
	rep.length = (strlen(nameList) + 1 + 3) >> 2;
	rep.sequenceNumber = client->sequence;
	if (client->swapped) {
		swaps(&rep.sequenceNumber, n);
		swapl(&rep.length, n);
	}
	WriteToClient(client, sizeof(xKACLGetUsersReply), (char *) &rep);
	if (rep.length)
		WriteToClient(client, strlen(nameList)+1, nameList);
	xfree(nameList);

	return (client->noClientException);
}

static int
ProcKACLDispatch(client)
register ClientPtr client;
{
	REQUEST(xReq);
	switch (stuff->data) {
	case X_KACLChangeUsers:
		return ProcKACLChangeUsers(client);
	case X_KACLGetUsers:
		return ProcKACLGetUsers(client);
	default:
		return BadRequest;
	}
}

static int
SProcKACLChangeUsers(client)
register ClientPtr client;
{
	register int    n;
	REQUEST(xKACLChangeUsersReq);

	swaps(&stuff->length, n);
	return ProcKACLChangeUsers(client);
}

static int
SProcKACLGetUsers(client)
register ClientPtr client;
{
	register int    n;
	REQUEST(xKACLGetUsersReq);

	swaps(&stuff->length, n);
	return ProcKACLGetUsers(client);
}

static int
SProcKACLDispatch(client)
register ClientPtr client;
{
	REQUEST(xReq);
	switch (stuff->data) {
	case X_KACLChangeUsers:
		return SProcKACLChangeUsers(client);
	case X_KACLGetUsers:
		return SProcKACLGetUsers(client);
	default:
		return BadRequest;
	}
}

/* 
 * add a user to the access control list. assumes that name is a null 
 * terminated string. sticks name at front of linked list.
#ifdef notdef
 * if this is the first name entered, and we are using kerberos (HASKRB
 * is defined) then we kill all currently connected clients for added
 * security.
#endif
 */
static int
AddUser (client, name)
ClientPtr client;
char *name;
{
	register ACLEntry *cur;
	int len;
	extern Bool AuthorizedClient();
#ifdef notdef
#ifdef HASKRB
	Bool zapOldClients = (listOfNames == NULL);
	int i;
#endif
#endif

	if (!AuthorizedClient(client))
		return BadAccess;

	for (cur = listOfNames; cur; cur = cur->next) {
		if (strcmp(cur->name,name) == 0)
			return Success;	/* already in there */
	}

 	cur = (ACLEntry *) xalloc(sizeof (ACLEntry));
	if (!cur)
		return BadAlloc;

	len = strlen(name) + 1;
	cur->name = (char *) xalloc(len);
	if (!cur->name) return BadAlloc;

	strcpy(cur->name, name);
	cur->next = listOfNames;
	listOfNames = cur;

#ifdef notdef
/* this doesn't seem to work -- it also resets the server */
#ifdef HASKRB
	if (zapOldClients)
		for (i = 1; i < currentMaxClients; ++i)
			if (clients[i])
				CloseDownClient(clients[i]);     
#endif
#endif

	return Success;
}

/* 
 * remove a user from the access control list. name is assumed to be a 
 * null terminated string.
 */
static int
DelUser (client, name)
ClientPtr client;
char *name;
{
	int len;
	register ACLEntry *cur, **prev;
	extern Bool AuthorizedClient();

	if (!AuthorizedClient(client))
		return BadAccess;

	for (prev = &listOfNames;
	     (cur = *prev) && (strcmp(cur->name,name));
	     prev = &cur->next)
		/* empty loop */;

	if (cur) {
		*prev = cur->next;
		xfree(cur);
	}

	return Success;
}

/* 
 * generate a list of all users in access control list as a 
 * colon-separated null-terminated string.  user is responsible for 
 * freeing the memory.
 */
static char *
GetUserList()
{
	char *list = NULL;
	char *newlist;
	int len = 0;
	register ACLEntry *cur;
	
	for (cur = listOfNames; cur; cur = cur->next) {
		len += strlen(cur->name) + 1;
		newlist = (char *)xrealloc(list,len+1);
		if (!newlist) {
			xfree(list);
			return NULL;
		}
		if (list == NULL) newlist[0] = '\0';	/* first element */
		list = newlist;
		strcat(list, cur->name);
		if (cur->next) strcat(list, ":");
	}

	return list;
}

/* 
 * the following two routines are internal interfaces to the acl for 
 * use from the kerberos authorization mechanism.  they allow the 
 * server to query the status of the ACL and determine if a particular 
 * item is in the list
 */

/* returns true if list is empty */
int
KACLEmptyList()
{
	return listOfNames == NULL;
}

/* 
 * returns true if char string is in the acl. assumes string is 
 * null-terminated like data in the list.
 */
int
KACLInList(data)
char *data;
{
	register ACLEntry *cur;

	for (cur = listOfNames; cur; cur = cur->next) {
		if (strcmp(cur->name,data) == 0)
			return 1;	/* its in there */
	}
	return 0;
}
