/*
 * Copyright 1990 MCNC and Vivek Khera
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of MCNC not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  MCNC makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * MCNC DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCNC
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Vivek Khera, MCNC
 */

/* 
 * library interface into the KACL extension
 * ideas mercilessly taken from MITMisc.c and the routines for Add/Get FontPath
 * V. Khera 26-JUL-1990
 * $Header: Kacl.c,v 1.4 90/09/18 10:26:41 bin Exp $
 */
#define NEED_REPLIES
#include "Xlibint.h"
#include "Kacl.h"
#include "kaclstr.h"
#include "Xext.h"
#include "extutil.h"

static XExtensionInfo *kacl_info;
static char kacl_extension_name[] = KACLNAME;

#define KACLCheckExtension(dpy,i,val) \
  XextCheckExtension (dpy, i, kacl_extension_name, val)

/*****************************************************************************
 *                                                                           *
 *			   private utility routines                          *
 *                                                                           *
 *****************************************************************************/

static int close_display();
static XExtensionHooks kacl_extension_hooks = {
    NULL,				/* create_gc */
    NULL,				/* copy_gc */
    NULL,				/* flush_gc */
    NULL,				/* free_gc */
    NULL,				/* create_font */
    NULL,				/* free_font */
    close_display,			/* close_display */
    NULL,				/* wire_to_event */
    NULL,				/* event_to_wire */
    NULL,				/* error */
    NULL				/* error_string */
};

static XEXT_GENERATE_FIND_DISPLAY (find_display,kacl_info,kacl_extension_name, 
				   &kacl_extension_hooks, KACLNumberEvents,
				   NULL)

static XEXT_GENERATE_CLOSE_DISPLAY (close_display, kacl_info)

/* 
 * public interface to access control list extension
 */

#if NeedFunctionPrototypes
Bool XKACLQueryExtension (
    Display *dpy,
    int *event_basep, int *error_basep)
#else
Bool XKACLQueryExtension (dpy, event_basep, error_basep)
    Display *dpy;
    int *event_basep, *error_basep;
#endif
{
    XExtDisplayInfo *info = find_display (dpy);

    if (XextHasExtension(info)) {
	*event_basep = info->codes->first_event;
	*error_basep = info->codes->first_error;
	return True;
    } else {
	return False;
    }
}

/* 
 * add a user to the access control list. assumes that user is 
 * null-terminated string containing one name.
 */
#if NeedFunctionPrototypes
Status XKACLAddUser(Display *dpy, const char *user)
#else
Status XKACLAddUser(dpy, user)
Display *dpy;
char *user;
#endif
{
	XExtDisplayInfo *info = find_display (dpy);
	register xKACLChangeUsersReq *req;
	int len, nbytes;
	char *p;

	KACLCheckExtension (dpy, info, 0);

	if (!user) return 0;

	LockDisplay(dpy);
	GetReq(KACLChangeUsers, req);
	len = strlen(user) + 1 + 1;	/* number of bytes + length byte */
	req->reqType = info->codes->major_opcode;
	req->aclReqType = X_KACLChangeUsers;
	req->mode = aclUserInsert;
	nbytes = (len + 3) & ~3;	/* align on word */
	BufAlloc(char *, p, nbytes);
	*p++ = len - 1;	/* length byte */
	strcpy(p, user);
	req->length += nbytes >> 2;
	UnlockDisplay(dpy);
	SyncHandle();
	return 1;
}

#if NeedFunctionPrototypes
Status XKACLDeleteUser(Display *dpy, const char *user)
#else
Status XKACLDeleteUser(dpy, user)
Display *dpy;
char *user;
#endif
{
	XExtDisplayInfo *info = find_display (dpy);
	register xKACLChangeUsersReq *req;
	int len, nbytes;
	char *p;

	KACLCheckExtension (dpy, info, 0);

	if (!user) return 0;

	LockDisplay(dpy);
	GetReq(KACLChangeUsers, req);
	len = strlen(user) + 1 + 1;	/* number of bytes + length byte */
	req->reqType = info->codes->major_opcode;
	req->aclReqType = X_KACLChangeUsers;
	req->mode = aclUserDelete;
	nbytes = (len + 3) & ~3;	/* align on word */
	BufAlloc(char *, p, nbytes);
	*p++ = len - 1;	/* length byte */
	strcpy(p, user);
	req->length += nbytes >> 2;
	UnlockDisplay(dpy);
	SyncHandle();
	return 1;
}

#if NeedFunctionPrototypes
char *XKACLGetUsers(Display *dpy)
#else
char *XKACLGetUsers(dpy)
Display *dpy;
#endif
{
	XExtDisplayInfo *info = find_display (dpy);
	register xKACLGetUsersReq *req;
	xKACLGetUsersReply rep;
	char *p = NULL;
	long nbytes;

	KACLCheckExtension (dpy, info, 0);

	LockDisplay(dpy);
	GetReq(KACLGetUsers, req);
	req->reqType = info->codes->major_opcode;
	req->aclReqType = X_KACLGetUsers;
	if (!_XReply(dpy, (xReply *)&rep, 0, xFalse)) {
		UnlockDisplay(dpy);
		SyncHandle();
		return NULL;
	}

	nbytes = (long)rep.length << 2;
	p = (char *) Xmalloc ((unsigned) nbytes);

	if (!p) {
		_XEatData(dpy, (unsigned long) nbytes);
		UnlockDisplay(dpy);
		SyncHandle();
		return (char *) NULL;
	}

	_XReadPad (dpy, p, nbytes);

	UnlockDisplay(dpy);
	SyncHandle();
	return p;
}

#if NeedFunctionPrototypes
void XKACLFreeList(const char *list)
#else
void XKACLFreeList(list)
char *list;
#endif
{
	Xfree(list);
}
